import logging
import json
from datetime import datetime, timedelta
from typing import Dict, Any
from flask import current_app
import requests


def main() -> str:
    """Harvest daily air quality data from GeoServer WFS.

    Handles:
    - Time window calculation (previous UTC day)
    - CQL filter construction for Mildura station
    - WFS API query
    - Data routing via Fission enqueue service

    Returns:
        'OK' with HTTP 200 on successful harvest

    Raises:
        requests.exceptions.RequestException: For API call failures
        json.JSONDecodeError: If invalid response received
    """
    # Calculate yesterday's date in UTC
    now: datetime = datetime.utcnow() - timedelta(days=1)
    
    # Build time range filter
    start: datetime = now.replace(hour=0, minute=0, second=0, microsecond=0)
    end: datetime = now.replace(hour=23, minute=59, second=59, microsecond=999999)
    cql_filter: str = f"site_name='Mildura' AND time_stamp>={start.isoformat()}Z AND time_stamp<={end.isoformat()}Z"

    # Configure WFS request parameters
    params: Dict[str, str] = {
        'service': 'WFS',
        'version': '2.0.0',
        'request': 'GetFeature',
        'typeName': 'geonode:vic_observations_2023',
        'outputFormat': 'application/json',
        'cql_filter': cql_filter
    }

    # Execute GeoServer request
    response = requests.get(
        url="https://naqd.eresearch.unimelb.edu.au/geoserver/wfs",
        params=params,
        verify=False,
        timeout=10
    )
    response.raise_for_status()

    # Log harvest metrics
    current_app.logger.info(
        f'Harvested air quality data for {start.date()} '
        f'({len(response.json()["features"])} records)'
    )

    # Route to message queue
    requests.post(
        url='http://router.fission/enqueue/airquality',
        headers={'Content-Type': 'application/json'},
        data=json.dumps(response.json()),
        timeout=5
    )

    return 'OK'
