import unittest
import requests
import json
from typing import Any
from unittest.mock import patch

class HTTPSession:
    """HTTP client session wrapper for making API requests."""
    
    def __init__(self, protocol: str, hostname: str, port: int) -> None:
        self.session = requests.Session()
        self.base_url = f'{protocol}://{hostname}:{port}'

    def get(self, path: str) -> requests.Response:
        """Send GET request to specified path."""
        return self.session.get(f'{self.base_url}/{path}')

    def post(self, path: str, data: dict[str, Any]) -> requests.Response:
        """Send POST request with JSON data to specified path."""
        return self.session.post(f'{self.base_url}/{path}', json=data)

    def put(self, path: str, data: dict[str, Any]) -> requests.Response:
        """Send PUT request with JSON data to specified path."""
        return self.session.put(f'{self.base_url}/{path}', json=data)

    def delete(self, path: str) -> requests.Response:
        """Send DELETE request to specified path."""
        return self.session.delete(f'{self.base_url}/{path}')

class TestEnd2End(unittest.TestCase):
    """End-to-end tests for student/course management API."""
    
    def test_student(self) -> None:
        """Verify student resource CRUD operations return valid JSON responses."""
        self.assertEqual(test_request.get('/students/111').status_code, 200)
        self.assertIsNotNone(test_request.get('/students/111').json())

        self.assertEqual(test_request.put('/students/111', {}).status_code, 200)
        self.assertIsNotNone(test_request.put('/students/111', {}).json())

        self.assertEqual(test_request.delete('/students/111').status_code, 200)
        self.assertIsNotNone(test_request.delete('/students/111').json())

    def test_students(self) -> None:
        """Verify students collection endpoint returns valid JSON response."""
        self.assertEqual(test_request.get('/students').status_code, 200)
        self.assertIsNotNone(test_request.get('/students').json())

    def test_course(self) -> None:
        """Verify course resource CRUD operations return valid JSON responses."""
        self.assertEqual(test_request.get('/courses/111').status_code, 200)
        self.assertIsNotNone(test_request.get('/courses/111').json())

        self.assertEqual(test_request.put('/courses/111', {}).status_code, 200)
        self.assertIsNotNone(test_request.put('/courses/111', {}).json())

        self.assertEqual(test_request.delete('/courses/111').status_code, 200)
        self.assertIsNotNone(test_request.delete('/courses/111').json())

    def test_courses(self) -> None:
        """Verify courses collection and related endpoints return valid JSON."""
        self.assertEqual(test_request.get('/courses').status_code, 200)
        self.assertIsNotNone(test_request.get('/courses').json())

        self.assertEqual(test_request.get('/courses/111/students').status_code, 200)
        self.assertIsNotNone(test_request.get('/courses/111/students').json())

if __name__ == '__main__':

    test_request = HTTPSession('http', 'localhost', 9090)
    unittest.main()
