from flask import Request, current_app
import requests, logging, json
from datetime import datetime
from typing import Dict, Any, Tuple, Optional

def config(k: str) -> str:
    """Read configuration parameter from default configs.
    
    Args:
        k: Configuration key filename to read
        
    Returns:
        String contents of the configuration file
    """
    with open(f'/configs/default/parameters/{k}', 'r') as f:
        return f.read()


def docurl() -> str:
    """Construct Elasticsearch document URL using course ID from headers."""
    return f'{config("ES_URL")}/{config("ES_DATABASE")}/_doc/course_{request.headers["X-Fission-Params-Courseid"]}'


def addfields(data: Dict[str, Any]) -> Dict[str, Any]:
    """Enrich course data with metadata.
    
    Adds:
        - timestamp: ISO format current datetime
        - id: Course ID from request headers
        - type: Hardcoded 'course' identifier
    """
    data['timestamp'] = datetime.now().isoformat()
    data['id'] = request.headers['X-Fission-Params-Courseid']
    data['type'] = 'course'
    return data


def main() -> Tuple[Optional[Dict[str, Any]], int]:
    """Handle course CRUD operations.
    
    Returns:
        Tuple containing:
        - Parsed JSON response (None on error)
        - HTTP status code
    """
    try:
        if request.method == 'GET':
            r: requests.Response = requests.get(
                docurl(),
                verify=False,
                auth=(config("ES_USERNAME"), config("ES_PASSWORD"))
            )
            return r.json(), r.status_code

        elif request.method == 'PUT':
            r = requests.get(docurl(), verify=False, 
                auth=(config("ES_USERNAME"), config("ES_PASSWORD")))
            
            params: Dict[str, int] = {}
            if r.status_code == 200:
                params = {
                    'if_seq_no': r.json()['_seq_no'],
                    'if_primary_term': r.json()['_primary_term']
                }

            r = requests.put(
                docurl(),
                verify=False,
                auth=(config("ES_USERNAME"), config("ES_PASSWORD")),
                headers={'Content-type': 'application/json'},
                data=json.dumps(addfields(request.get_json(silent=True) or {})),
                params=params
            )
            return r.json(), r.status_code

        elif request.method == 'DELETE':
            r = requests.get(docurl(), verify=False,
                auth=(config("ES_USERNAME"), config("ES_PASSWORD")))
            
            if r.status_code != 200:
                return r.json(), r.status_code

            r = requests.delete(
                docurl(),
                verify=False,
                auth=(config("ES_USERNAME"), config("ES_PASSWORD")),
                params={
                    'if_seq_no': r.json()['_seq_no'],
                    'if_primary_term': r.json()['_primary_term']
                }
            )
            return r.json(), r.status_code

    except Exception as e:
        current_app.logger.error(str(e))
        return {'message': f'Error: {str(e)}'}, 500

    return {'message': 'Method not allowed'}, 405
