import unittest
import requests
import json
import time
from typing import Dict, Any, Union, List

class HTTPSession:
    """HTTP client wrapper for API testing.
    
    Attributes:
        base_url: Base URL for all requests (protocol://host:port)
    """
    def __init__(self, protocol: str, hostname: str, port: int) -> None:
        """Initialize HTTPSession instance.

        Args:
            protocol: Protocol for the base URL (e.g., http, https)
            hostname: Hostname for the base URL
            port: Port number for the base URL
        """
        self.session = requests.session()
        self.base_url = f'{protocol}://{hostname}:{port}'

    def get(self, path: str) -> requests.Response:
        """Send GET request to specified path.

        Args:
            path: Path for the GET request

        Returns:
            Response object from the GET request
        """
        return self.session.get(f'{self.base_url}{path}')

    def post(self, path: str, data: Union[Dict[str, Any], List[Any]]) -> requests.Response:
        """Send POST request with JSON payload.

        Args:
            path: Path for the POST request
            data: JSON payload for the POST request

        Returns:
            Response object from the POST request
        """
        return self.session.post(f'{self.base_url}{path}', json=data)

    def put(self, path: str, data: Union[Dict[str, Any], List[Any]]) -> requests.Response:
        """Send PUT request with JSON payload.

        Args:
            path: Path for the PUT request
            data: JSON payload for the PUT request

        Returns:
            Response object from the PUT request
        """
        return self.session.put(f'{self.base_url}{path}', json=data)

    def delete(self, path: str) -> requests.Response:
        """Send DELETE request.

        Args:
            path: Path for the DELETE request

        Returns:
            Response object from the DELETE request
        """
        return self.session.delete(f'{self.base_url}{path}')


class TestEnd2End(unittest.TestCase):
    """End-to-end test suite for student/course management API."""

    def setUp(self) -> None:
        """Reset database before each test."""
        self.assertEqual(test_request.delete('/wipedatabase').status_code, 200)
        time.sleep(1)

    def test_student(self) -> None:
        """Test student CRUD operations."""
        # Create students
        self.assertEqual(test_request.put('/students/1', {'name': 'John Doe', 'courses': '90024'}).status_code, 201)
        self.assertEqual(test_request.put('/students/2', {'name': 'Jane Doe', 'courses': ['90024', '90059']}).status_code, 201)
        time.sleep(1)

        # Verify initial data
        r = test_request.get('/students/1')
        self.assertEqual(r.status_code, 200)
        o: Dict[str, Any] = r.json()['_source']
        self.assertEqual(o['name'], 'John Doe')
        self.assertEqual(o['courses'], '90024')

        r = test_request.get('/students/2')
        o = r.json()['_source']
        self.assertEqual(r.status_code, 200)
        self.assertEqual(o['name'], 'Jane Doe')
        self.assertEqual(o['courses'], ['90024', '90059'])

        # Test update
        r = test_request.put('/students/1', {'name': 'Bob Carr', 'courses': '90024'})
        self.assertEqual(r.status_code, 200)
        r = test_request.get('/students/1')
        o = r.json()['_source']
        self.assertEqual(r.status_code, 200)
        self.assertEqual(o['name'], 'Bob Carr')
        self.assertEqual(o['courses'], '90024')

        # Test not found
        self.assertEqual(test_request.get('/students/999').status_code, 404)

        # Cleanup
        self.assertEqual(test_request.delete('/students/1').status_code, 200)
        self.assertEqual(test_request.delete('/students/2').status_code, 200)

    def test_students(self) -> None:
        """Test course-student relationships."""
        # Setup courses
        self.assertEqual(test_request.put('/courses/90024', {'name': 'Cloud Computing'}).status_code, 201)
        self.assertEqual(test_request.put('/courses/90059', {'name': 'Introduction to Programming'}).status_code, 201)
        
        # Create student
        self.assertEqual(test_request.put('/students/1', {'name': 'John Doe', 'courses': '90024'}).status_code, 201)

    def test_course(self) -> None:
        """Test course CRUD operations."""
        self.assertEqual(test_request.put('/courses/90024', {'name': 'Cloud x Computing'}).status_code, 201)
        self.assertEqual(test_request.put('/courses/90059', {'name': 'Introduction to Programming'}).status_code, 201)
        time.sleep(1)

        r = test_request.get('/courses/90024')
        o = r.json()['_source']
        self.assertEqual(r.status_code, 200)
        self.assertEqual(o['name'], 'Cloud x Computing')

        r = test_request.get('/courses/90059')
        o = r.json()['_source']
        self.assertEqual(r.status_code, 200)
        self.assertEqual(o['name'], 'Introduction to Programming')

        r = test_request.put('/courses/90024', {'name': 'Cloud Computing'})
        self.assertEqual(r.status_code, 200)
        r = test_request.get('/courses/90024')
        o = r.json()['_source']
        self.assertEqual(r.status_code, 200)
        self.assertEqual(o['name'], 'Cloud Computing')

        self.assertEqual(test_request.get('/courses/999').status_code, 404)

        self.assertEqual(test_request.delete('/courses/90024').status_code, 200)
        self.assertEqual(test_request.delete('/courses/90059').status_code, 200)

    def test_courses(self) -> None:
        """Test course listing."""
        self.assertEqual(test_request.put('/courses/90024', {'name': 'Cloud Computing'}).status_code, 201)
        self.assertEqual(test_request.put('/courses/90059', {'name': 'Introduction to Programming'}).status_code, 201)
        time.sleep(1)

        r = test_request.get('/courses')
        o = (r.json()['hits'])['hits']
        self.assertEqual(r.status_code, 200)
        self.assertEqual(o[0]['fields']['name'][0], 'Cloud Computing')
        self.assertEqual(o[1]['fields']['name'][0], 'Introduction to Programming')

    def test_courses_students(self) -> None:
        """Test course-student relationships."""
        self.assertEqual(test_request.put('/courses/90024', {'name': 'Cloud Computing'}).status_code, 201)
        self.assertEqual(test_request.put('/courses/90059', {'name': 'Introduction to Programming'}).status_code, 201)
        self.assertEqual(test_request.put('/students/1', {'name': 'John Doe', 'courses': '90024'}).status_code, 201)
        self.assertEqual(test_request.put('/students/2', {'name': 'Jane Doe', 'courses': ['90024', '90059']}).status_code, 201)
        time.sleep(1)

        r = test_request.get('/courses/90024/students')
        o = (r.json()['hits'])['hits']
        self.assertEqual(r.status_code, 200)
        self.assertEqual(len(o), 2)
        self.assertEqual(o[0]['fields']['name'][0], 'Jane Doe')
        self.assertEqual(o[1]['fields']['name'][0], 'John Doe')

        r = test_request.get('/courses/90059/students')
        o = (r.json()['hits'])['hits']
        self.assertEqual(len(o), 1)
        self.assertEqual(r.status_code, 200)
        self.assertEqual(o[0]['fields']['name'][0], 'Jane Doe')

        r = test_request.get('/courses/99999/students')
        o = (r.json()['hits'])['hits']
        self.assertEqual(r.status_code, 200)
        self.assertEqual(len(o), 0)

if __name__ == '__main__':
    test_request = HTTPSession('http', 'localhost', 9090)
    unittest.main()
